"""
Dead Hosts's launcher - The launcher of the Dead-Hosts infrastructure.

Provides a simple implementation of the python :code:`re` package.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Project link:
    https://github.com/dead-hosts/infrastructure-launcher

License:
::

    MIT License

    Copyright (c) 2019 Dead Hosts
    Copyright (c) 2019 Nissar Chababy

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

from re import compile as comp
from re import escape as esc
from re import sub


class Regex:
    """
    Provides a simple implementation of the python :code:`re` package.

    :param data: The data to work with.
    :type data: str, list
    :param str regex: The regex to work with.
    :param bool escape: Should we escape the given :code:`regex`?

    :ivar regex: The regex we work with.
    :ivar data: The data we work with.
    """

    def __init__(self, data, regex, escape=False):
        self.data = data

        if escape:
            self.regex = esc(regex)
        else:
            self.regex = regex

    def get_not_matching_list(self):
        """
        Returns a list of string which don't match the given :code:`regex`.

        .. warning::
            We expect :code:`self.data` to be a list of string.
        """

        pre_result = comp(self.regex)

        return [x for x in self.data if not pre_result.search(x)]

    def get_matching_list(self):
        """
        Returns a list of string which match the given :code:`regex`.

        .. warning::
            We expect :code:`self.data` to be a list of string.
        """

        pre_result = comp(self.regex)

        return [x for x in self.data if pre_result.search(x)]

    def match(self, rematch=False, return_data=True, group=0):
        """
        Checks if the given :code:`data` matches the given :code:`regex`.

        :param bool rematch:
            Return the matched group into a formatted list.

            .. note::
                This allows us to have an implementation of the Bash :code:`${BASH_REMATCH}`.
        :param bool return_data:
            Tell us to return what was matched.
        :param int group:
            The index of the group to return.

        :rtype: str, list, bool
        """

        to_match = comp(self.regex)

        if rematch:
            pre_result = to_match.findall(self.data)
        else:
            pre_result = to_match.search(self.data)

        if return_data and pre_result:
            if rematch:
                result = []

                for data in pre_result:
                    if isinstance(data, tuple):
                        result.extend(list(data))
                    else:
                        result.append(data)

                if group != 0:
                    return result[group]
                return result
            return pre_result.group(group).strip()

        if not return_data and pre_result:
            return True
        return False

    def replace(self, replacement, occurences=0):
        """
        Replaces the matched part with the given :code:`replacement`.

        :param str replacement: The replacement.
        :param int occurences:
            The number of occurences to replace.

            .. note::
                If :code:`0` is given, we replace every occurences.

        :rtype: str
        """

        if replacement:
            return sub(self.regex, replacement, self.data, occurences)
        return self.data



"""
Dead Hosts's launcher - The launcher of the Dead-Hosts infrastructure.

Provides the update logic. Basically, the update of all files about
the repository, the infrastructure or project.Is

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Project link:
    https://github.com/dead-hosts/infrastructure-launcher

License:
::

    MIT License

    Copyright (c) 2019 Dead Hosts
    Copyright (c) 2019 Nissar Chababy

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

import logging

import PyFunceble.helpers as helpers

from ..configuration import Links, Markers, Paths
from ..configuration import PyFunceble as PyFuncebleConfig
from ..configuration import TravisCI as TravisCIConfig


class Update:
    """
    Provides an interface for the updates.

    :param str working_directory: The directory we are working from.

    :ivar working_directory: Same as :code:`working_directory`.
    """

    def __init__(self, working_directory, info_manager):
        self.working_directory = working_directory
        self.info_manager = info_manager

        self.cross_pyfunceble_config_file = helpers.File(
            self.working_directory
            + Links.cross_pyfunceble_configuration["link"].split("/")[-1]
        )
        self.pyfunceble_config_file = helpers.File(
            self.working_directory + Links.cross_pyfunceble_configuration["destination"]
        )

        logging.info(
            f"Cross repository PyFunceble config file: {self.cross_pyfunceble_config_file.path}"
        )
        logging.info(f"Local config file: {self.pyfunceble_config_file.path}")

        self.pyfunceble_config()
        self.pyfunceble_official_config()
        self.pyfunceble_official_license()

        self.readme()
        self.our_license()
        self.travis_yml()

    def get_final_commit_message(self):
        """
        Provides the final and update commit message.
        """

        if self.info_manager.ping:
            return (
                PyFuncebleConfig.travis_autosave_final_commit
                + f" | cc {self.info_manager.get_ping_for_commit()} | "
            )
        return PyFuncebleConfig.travis_autosave_final_commit

    def __cross_pyfunceble_config(self):
        """
        Updates the cross configuration of PyFunceble.
        """

        if self.cross_pyfunceble_config_file.exists() or not TravisCIConfig.git_email:
            logging.info("Cross repository file exists locally. Let's update it.")

            upstream_string_version = helpers.Download(
                Links.pyfunceble_official_config["link"]
            ).text()

            local_version = helpers.Dict.from_yaml(upstream_string_version)
            local_version = helpers.Merge(
                {
                    x: y
                    for x, y in PyFuncebleConfig.__dict__.items()
                    if not x.startswith("_")
                }
            ).into(local_version, strict=True)

            local_version_2 = local_version.copy()

            if self.info_manager.ping:
                logging.info("Ping names given, appending them to the commit message.")

                local_version_2[
                    "travis_autosave_final_commit"
                ] = self.get_final_commit_message()

            if self.info_manager.custom_pyfunceble_config and isinstance(
                self.info_manager.custom_pyfunceble_config, dict
            ):
                logging.info(
                    "Custom PyFunceble configuration given, "
                    "appending them to the local configuration file."
                )
                local_version_2 = helpers.Merge(
                    self.info_manager.custom_pyfunceble_config
                ).into(local_version_2, strict=True)

            helpers.Dict(local_version).to_yaml_file(
                self.cross_pyfunceble_config_file.path
            )
            helpers.Dict(local_version_2).to_yaml_file(self.pyfunceble_config_file.path)
            logging.info(
                "Finished update of the cross repository PyFunceble config file."
            )
            logging.info("Updated the local PyFunceble configuration too.")

    def pyfunceble_config(self):
        """
        Updates the PyFunceble cross repository configuration.
        """

        self.__cross_pyfunceble_config()

        if (
            not self.info_manager.own_management
            and not self.cross_pyfunceble_config_file.exists()
        ):
            logging.info(
                "Cross repository file dooesn't exists locally. "
                "Let's fetch it into our local PyFunceble config."
            )

            cross_repository_version = helpers.Download(
                Links.cross_pyfunceble_configuration["link"]
            ).text()

            local_version = helpers.Dict().from_yaml(cross_repository_version)

            if self.info_manager.ping:
                local_version[
                    "travis_autosave_final_commit"
                ] = self.get_final_commit_message()
                logging.info("Ping names given, appending them to the commit message.")

            if self.info_manager.custom_pyfunceble_config and isinstance(
                self.info_manager.custom_pyfunceble_config, dict
            ):
                logging.info(
                    "Custom PyFunceble configuration given, "
                    "appending them to the local configuration file."
                )
                local_version = helpers.Merge(
                    self.info_manager.custom_pyfunceble_config
                ).into(local_version, strict=True)

            logging.debug(f"Interpreded PyFunceble config: {local_version}")

            helpers.Dict(local_version).to_yaml_file(self.pyfunceble_config_file.path)

    def readme(self):
        """
        Updates the README file.
        """

        destination_file_instance = helpers.File(
            self.working_directory + Paths.readme_filename
        )

        if destination_file_instance.exists():
            logging.info(
                f"{destination_file_instance.path} exists, let's update it's content."
            )

            updated_version = helpers.Regex(
                Markers.extract_about_pyfunceble
            ).replace_match(destination_file_instance.read(), Markers.about_pyfunceble)
            logging.info(
                f"Updated `About PyFunceble` section of {destination_file_instance.path}"
            )

            updated_version = helpers.Regex(
                Markers.extract_about_dead_hosts
            ).replace_match(updated_version, Markers.about_dead_hosts)
            logging.info(
                f"Updated `About Dead-Hosts` section of {destination_file_instance.path}"
            )

            destination_file_instance.write(updated_version, overwrite=True)

    def our_license(self):
        """
        Updates the local version of our LICENSE file.
        """

        helpers.Download(Links.our_license["link"]).text(
            destination=self.working_directory + Links.our_license["destination"]
        )
        logging.info(
            f"Updated {self.working_directory + Links.our_license['destination']}"
        )

    def pyfunceble_official_config(self):
        """
        Update the official version of the PyFunceble file.
        """

        helpers.Download(Links.pyfunceble_official_config["link"]).text(
            destination=self.working_directory
            + Links.pyfunceble_official_config["destination"]
        )

        logging.info(
            f"Updated {self.working_directory +  Links.pyfunceble_official_config['destination']}"
        )

    def pyfunceble_official_license(self):
        """
        Update the official version of the PyFunceble license.
        """

        helpers.Download(Links.pyfunceble_official_license["link"]).text(
            destination=self.working_directory
            + Links.pyfunceble_official_license["destination"]
        )

        logging.info(
            f"Updated {self.working_directory +  Links.pyfunceble_official_license['destination']}"
        )

    def travis_yml(self):
        """
        Updates the `.travis.yml` file.
        """

        if (
            TravisCIConfig.build_dir
            and TravisCIConfig.github_token
            and not helpers.File(self.working_directory + "info.example.json").exists()
        ):

            to_delete_main = ["dist", "cache", "matrix", "python"]
            to_delete_global_end = ["UPDATE_ME_LOCATION", "ADMIN_LOCATION"]

            destination = self.working_directory + Paths.travis_filename
            destination_file_instance = helpers.File(destination)

            logging.info(f"Updating {destination_file_instance.path}.")
            logging.debug(
                f"{destination_file_instance.path} exists: {destination_file_instance.exists()}"
            )

            content = helpers.Dict().from_yaml(destination_file_instance.read())
            content = helpers.Merge(TravisCIConfig.unified_config).into(
                content, strict=True
            )

            for index in to_delete_main:
                if index in content:
                    del content[index]

            if "env" in content and "global" in content["env"]:
                for index, data in enumerate(content["env"]["global"]):
                    for env_var in to_delete_global_end:
                        if env_var in data:
                            del content["env"]["global"][index]

            to_write = helpers.Dict(content).to_yaml()

            logging.debug(
                f"New repository TravisCI configuration (interpreted): \n{content}"
            )
            destination_file_instance.write(to_write, overwrite=True)



"""
Dead Hosts's launcher - The launcher of the Dead-Hosts infrastructure.

Provides the authorization logic.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Project link:
    https://github.com/dead-hosts/infrastructure-launcher

License:
::

    MIT License

    Copyright (c) 2019 Dead Hosts
    Copyright (c) 2019 Nissar Chababy

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

from datetime import datetime, timedelta

import PyFunceble.helpers as helpers

from ..configuration import Markers


class Authorize:
    """
    Provides some authorization logic.

    :param info_manager: An instance of the info manager.

    :ivar info_manager: The info manager.
    """

    def __init__(self, info_manager):
        self.info_manager = info_manager

    @classmethod
    def is_launch_given_by_commit_message(cls):
        """
        Checks if the launch per commit message is valid.
        """

        return helpers.Regex(Markers.launch_test).match(
            helpers.Command("git log -1").execute(), return_match=False
        )

    def refresh(self):
        """
        Provides the refresh authorization.

        If this authorization is not given, we should not update
        modify any files related to the original list we are testing.

        :rtype: bool
        """

        if (
            not self.info_manager.currently_under_test
            or self.is_launch_given_by_commit_message()
        ):
            return True

        return False

    def get_test_authorization_time(self):
        """
        Provides the expected (minimal) date of authorization
        of the next build.
        """

        return self.info_manager.finish_datetime + timedelta(
            days=self.info_manager.days_until_next_test
        )

    def test(self):
        """
        Provides the test authorization.

        If this authorization is not given, we should not start
        the test.

        :rtype: bool
        """

        if self.is_launch_given_by_commit_message():
            return True

        if (
            not self.info_manager.days_until_next_test
            or self.info_manager.days_until_next_test < 0
        ):
            return True

        if datetime.now() > self.get_test_authorization_time():
            return True

        if self.info_manager.currently_under_test:
            return True

        return False



"""
Dead Hosts's launcher - The launcher of the Dead-Hosts infrastructure.

Provides a way to manipulate :code:`dict`.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Project link:
    https://github.com/dead-hosts/infrastructure-launcher

License:
::

    MIT License

    Copyright (c) 2019 Dead Hosts
    Copyright (c) 2019 Nissar Chababy

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

from io import TextIOWrapper
from json import decoder, dump, dumps, load, loads

from yaml import dump as yaml_dump
from yaml import safe_load as yaml_load

import dead_hosts.launcher.helpers as helpers


class Dict:
    """
    Provides a interface for dictionaries manipulation.

    :param dict main: The main dictionary.

    :ivar main: The givne :code:`main`.
    """

    def __init__(self, main=None):
        self.main = main

    def to_json(self, destination=None, encoding="utf-8"):
        """
        Converts :code:`self.main` into a JSON formatted string.

        :param str destination: A destination to save the JSON string.
        :param str encoding: The encoding to use if we save the JSON string into a file.

        :rtype: str, None
        """

        if destination:
            with open(destination, "w", encoding=encoding) as file_stream:
                return dump(
                    self.main, file_stream, ensure_ascii=False, indent=4, sort_keys=True
                )
        return dumps(self.main, ensure_ascii=False, indent=4, sort_keys=True)

    def to_yaml(self, destination=None, encoding="utf-8", flow_style=False):
        """
        Converts :code:`self.main` into a YAML formatted string.

        :param str destination: A destination to save the YAML formatted string.
        :param str encoding: The encoding to use.
        :param bool flow_style: Tell us to follow or not the default flow style.

        :rtype: str, NOne
        """

        if destination:
            with open(destination, "w", encoding=encoding) as file_stream:
                return yaml_dump(
                    self.main,
                    file_stream,
                    encoding=encoding,
                    allow_unicode=True,
                    indent=4,
                    default_flow_style=flow_style,
                )
        return yaml_dump(
            self.main,
            encoding=encoding,
            allow_unicode=True,
            indent=4,
            default_flow_style=flow_style,
        ).decode(encoding)

    @classmethod
    def from_json_string(cls, data):
        """
        Converts the given JSON string (:code:`data`) into a :code:`dict`.

        :param str data: The JSON string to convert to :code:`dict`.

        :rtype: dict
        """

        try:
            return loads(data)
        except decoder.JSONDecodeError:
            return {}

    @classmethod
    def from_json_file(cls, file, encoding="utf-8"):
        """
        Converts the given JSON file (:code:`file`) into a :code:`dict`.

        :param file: The JSON file to convert to :code:`dict`.
        :type file: TextIOWrapper, str

        :param str encoding: The encoding to use while opening the file path.

        :rtype: dict
        """

        if isinstance(file, TextIOWrapper):
            return load(file)

        with open(file, "r", encoding=encoding) as file_stream:
            return load(file_stream)

    @classmethod
    def from_yaml_string(cls, data):
        """
        Converts the given YAML string (:code:`data`) into a :code:`dict`.

        :param str data: The YAML string to convert to :code:`dict`.

        :rtype: dict
        """

        return yaml_load(data)

    @classmethod
    def from_yaml_file(cls, file, encoding="utf-8"):
        """
        Converts the given YAML file (:code:`file`) into a :code:`dict`.

        :param file: The YAML file to convert to :code:`dict`.
        :type file: TextIOWrapper, str

        :param str encoding: The encoding to use while opening the file path.

        :rtype: dict
        """

        if isinstance(file, TextIOWrapper):
            return yaml_load(file)

        with open(file, "r", encoding=encoding) as file_stream:
            return yaml_load(file_stream)

    def merge(self, to_merge, strict=True):
        """
        Merge the given :code:`to_merge` into the given :code:`main`.

        :param dict to_merge: The :code:`dict` to merge.
        :param bool strict: Tell us if we have to respect index value or not while mergin lists.

        :rtype: dict
        """

        result = {}

        for index, data in to_merge.items():
            if index in self.main:
                if isinstance(data, dict) and isinstance(self.main[index], dict):
                    result[index] = Dict(self.main[index]).merge(data, strict=strict)
                elif isinstance(data, list) and isinstance(self.main[index], list):
                    result[index] = helpers.list.List(self.main[index]).merge(
                        data, strict=strict
                    )
                else:
                    result[index] = data
            else:
                result[index] = data

        for index, data in self.main.items():
            if index not in result:
                result[index] = data

        return result



"""
Dead Hosts's launcher - The launcher of the Dead-Hosts infrastructure.

Provides a way to download things.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Project link:
    https://github.com/dead-hosts/infrastructure-launcher

License:
::

    MIT License

    Copyright (c) 2019 Dead Hosts
    Copyright (c) 2019 Nissar Chababy

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

import requests

from .file import File


class Download:
    """
    Provides an interface for file download.

    :param str link: The link to download.
    """

    def __init__(self, link):
        self.link = link

    def text(self, destination=False, success_status_code=None):
        """
        Downloads the text of the given link.

        :param str destination:
            Where we are supposed to write the downloaded content.

            .. note::
                If a non :code:`str` value is given, we don't save into a file.
        :param list success_status_code: The list of HTTP status code to consider as success.
        """

        if not success_status_code:
            success_status_code = [200]
        elif not isinstance(success_status_code, list):
            success_status_code = [success_status_code]

        req = requests.get(self.link)

        if req.status_code in success_status_code:
            if destination and isinstance(destination, str):
                File(destination).write(req.text, overwrite=True)

            return req.text

        raise ValueError(
            f"<status_code> ({req.status_code}) not in the <success_status_code> list."
        )



"""
Dead Hosts's launcher - The launcher of the Dead-Hosts infrastructure.

Provides a way to simply manipulate files.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Project link:
    https://github.com/dead-hosts/infrastructure-launcher

License:
::

    MIT License

    Copyright (c) 2019 Dead Hosts
    Copyright (c) 2019 Nissar Chababy

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

from os import path, remove


class File:
    """
    Provides an interface for file manipulation.

    :param str file: A path to the file to work with.
    :param str encoding: The encoding to use while reading or writting :code:`file`.

    :ivar file: The file we currently work with.
    :ivar encoding: The encoding to use.
    """

    def __init__(self, file, encoding="utf-8"):
        self.file = file
        self.encoding = encoding

    def exists(self):
        """
        Checks if the given :code:`self.file` exists.

        :rtype: bool
        """

        return path.isfile(self.file)

    def read(self):
        """
        Reads the given :code:`self.file`.
        """

        with open(self.file, "r", encoding=self.encoding) as file_stream:
            funilrys = file_stream.read()

        return funilrys

    def read_to_list(self):
        """
        Reads the given :code:`self.file` and convert each lines to a :code:`list`.
        """

        return self.read().splitlines()

    def write(self, data, overwrite=False):
        """
        Writes the given :code:`data` into the given :code:`self.file`.

        :param str data: The data to write.
        :param bool overwrite: Should we overwrite the file ?
        """

        if data and isinstance(data, str):
            if overwrite or not self.exists():
                mode = "w"
            else:
                mode = "a"

            with open(self.file, mode, encoding=self.encoding) as file_stream:
                file_stream.write(data)

    def delete(self):
        """
        Deletes the given :code:`self.file` (if exists).
        """

        if self.exists():
            remove(self.file)
